# Copyright (C) 2005 Mandrakesoft
#  
# Author: Florent Villard <warly@mandrakesoft.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#       
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#               
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Mkcd::Shell;

my $VERSION = '0.0.1';

use strict;

use Term::ReadLine;
use Mkcd::Commandline qw(parseCommandLine usage);
use Mkcd::Tools qw(printDiscsFile);

use base qw(Term::ReadLine);

my $shell;
my $exit;
my %functions = ( 
    packages => [ 
	[ "", "packages", 0, "[options]", "List and manipulate packages list", \&packages, "Manipulating packages lists" ],
	[ "l", "list", 
	   [ 
	       [ "", "list", -1, "<package 1> <package 2> ... <package 3>", "List and manipulate packages list", 
	           sub {
		      my ($tmp, @arg) = @_;
		      $tmp->[0] ||= {};
		      push @$tmp, @arg;
		      1
	           },
			"Manipulating packages list" ],
	       [ "e", "exclude", 0, "", "Exclude the given packages", 
	       		sub { my ($tmp) = @_; $tmp->[0]{exclude} = 1; 1}, 
			"Setting the exclude flag" ],
	       [ "r", "regexp", 0, "", "The packages name will be considered as regular expressions", 
	       		sub { my ($tmp) = @_; $tmp->[0]{regexp} = 1; 1}, 
			"Setting the regexp flag" ],
	       [ "e", "needed", 1, "<rep number>", "The packages will be put in or before given rep",
	       		sub { my ($tmp, $rep) = @_; $tmp->[0]{needed} = $rep; 1}, 
			"Setting the needed flag" ],
	   ], "[options] <list>", "List and manipulate packages list",
	         \&list,
		"Manipulating packages lists" ],
	[ "p", "print", 0, "", "Print current packages list and display it", \&packages_print, "Printing packages list" ],
    ],
    size => [ 
	[ "", "size", 0, "<options>", "Print various size informations", \&size, "Printing size data" ],
    ],
    status => [ 
	[ "", "status", 0, "", "Print current status", \&status, "Printing status" ],
    ],
    next => [ [ "", "next", 0, "", "Perform next iteration of building", sub { print "Next iteration\n"; $exit = 1 }, "Quitting the mkcd shell"  ] ],
    quit => [ [ "", "quit", 0, "", "Exit this mkcd session", sub { print "Exiting\n"; exit }, "Quitting the mkcd shell"  ] ],
    resume => [ [ "", "resume", 0, "", "Resume the mkcd session", sub { print "Resuming\n"; $exit = 1 }, "Quitting the mkcd shell"  ] ],
);


sub new {
    my ($invocant, @args) = @_;
    my $class = ref($invocant) || $invocant;

    my $self = $class->SUPER::new(@args);
    $shell = $self;
    $self->Attribs->{attempted_completion_function} = \&mkcd_completion;

    return $self;
}

sub prompt {
    my ($class, $config) = @_;
    $exit = 0;
    while (defined (my $line = $class->readline('mkcd> '))) {
	print "\n";	
	$line =~ /^$/ and next;
	my ($word, @arg) = split(' ', $line);
	unless ($functions{$word}) {
	    printf STDERR "$word: invalid command (type help for the list of available commands).\n";
	    next
	}
	my $todo = parseCommandLine($word, \@arg, $functions{$word}, 1);
	@arg and usage($word, $functions{$word}, "@arg, too many arguments", 1), next;
	foreach my $t (@$todo)  {
	    &{$t->[0]}($config, @{$t->[1]}) or print "error: $t->[2]\n";
	}
	$exit and last
    }
    1
}

sub status {
    my ($class) = @_;
    print "Status: OK\n"
}

sub mkcd_completion {
    my ($text, $line, $start, $end) = @_;

    my @matches = ();

    @matches = $shell->completion_matches ($text, \&command_generator) if ($start == 0);

    @matches
}

{
    my $list_index;
    my @name;

    sub command_generator {
	my ($text, $state) = @_;

	unless ($state) {
	    $list_index = 0;
	    @name = keys(%functions)
	}

	while ($list_index <= $#name) {
	    $list_index++;
	    return $name[$list_index - 1]
	    if ($name[$list_index - 1] =~ /^$text/)
	}
	undef
    }
}

sub size {
    my ($config) = @_;
    my $disc_size = $config->{size}{disc};
    my @size_name = ('b', 'kb', 'mb', 'gb', 'tb', 'pb');
    for (my $i; $i < @$disc_size; $i++) {
	$disc_size->[$i] or next;
	my $hs = $disc_size->[$i];
	my $j;
        while ($hs > 1000) { $j++; $hs = $hs >> 10 }
	print "Disc $i: $disc_size->[$i] ($hs $size_name[$j])\n"
    }
    my $disc_rep = $config->{size}{rep};
    foreach my $cd (keys %$disc_rep) {
	foreach my $rep (keys %{$disc_rep->{$cd}}) {
	    foreach my $list (keys %{$disc_rep->{$cd}{$rep}}) {
		my $hs = $disc_rep->{$cd}{$rep}{$list};
		my $j;
		while ($hs > 1000) { $j++; $hs = $hs >> 10 }
		print "Disc $cd rep $rep list $list: $disc_rep->{$cd}{$rep}{$list} ($hs $size_name[$j])\n"
	    }
	}
    }
    1
}

sub packages_print {
    my ($config) = @_;
    my $discs_files = $config->{discs_files};
    my $tmpdir = $config->{config}{tmp} || $ENV{TMPDIR} || '/tmp/';
    my $time = time;
    my $dir = "$tmpdir/mkcd-shell";
    -d $dir or mkdir $dir;
    my $file = "$dir/packages.$time";
    push @{$config->{written_files}}, $file;
    print "Writing $file\n";
    printDiscsFile($config->{config}, $config->{discs_files}, $file, $config->{metagroups});
    my $editor = $ENV{EDITOR} || 'vi';
    system($editor, $file);
    1
}

sub packages {
    my ($config) = @_;
    my $discs_files = $config->{discs_files};
    print "File written @{$config->{written_files}}\n" if ref $config->{written_files};
    for (my $cd; $cd < @$discs_files; $cd++) {
	ref $discs_files->[$cd] or next;
	foreach my $rep (keys %{$discs_files->[$cd]}) {
	    foreach my $list (sort keys %{$discs_files->[$cd]{$rep}}) {
		my $nbrpm = keys %{$discs_files->[$cd]{$rep}{$list}};
		print "Disc $cd rep $rep list $list: $nbrpm packages\n";
	    }
	}
    }
}

sub resume {
    my ($config) = @_;
    print "Resuning...\n";
    $exit = 1
}

sub list {
    my ($config, $options, @packages) = @_;
    print "List config\n";
    foreach my $k (keys %$options) {
	print "K $k\n";
    }
    print "PACKAGES @packages\n";
    1
}
1
